import copy
import re

from six import print_

from . import helper, requester
from .exceptions import TkStackException
from .helper import CONST, Group, Logger

_log = None


def complete(component, nodes_yaml, tosca,
             policy=None, constraints=None, interactive=False,
             df_host=CONST.DF_HOST):
    global _log
    _log = Logger.get(__name__)

    properties = component.constraints
    _log.debug('properties: {}'.format(properties))

    count, images = _get_images(properties, policy, constraints, df_host)

    print_('founded {:0} images for "{.name}" node'
           ''.format(count, component))

    if count == 0:
        print_('[WARNING] node "{.name}" will not be completed'
               ''.format(component))
        return False

    image = _choose_image(images, interactive)

    if isinstance(component, Group):
        _update_group_yaml(component, nodes_yaml, image)
        print_('complete group "{}" with image "{}" '
               '({:.2f} MB, {} pulls, {} stars)'
               ''.format(component.name, image['name'],
                         image['size'] / 1000000, image['pulls'],
                         image['stars']))
    else:
        _update_yaml(component, nodes_yaml, image)
        print_('complete node "{}" with image "{}" '
               '({:.2f} MB, {} pulls, {} stars)'
               ''.format(component.name, image['name'],
                         image['size'] / 1000000, image['pulls'],
                         image['stars']))
    return True


def _get_images(properties, policy=None, constraints=None,
                df_host=CONST.DF_HOST):
    query = _build_query(properties, policy, constraints)
    _log.debug('query {}'.format(query))

    count, images = requester.search_images(query, df_host)
    _log.debug('returned images..')

    if len(images) < 1:
        return 0, None
    return count, images


def _build_query(properties, policy=None, constraints={}):
    policy = CONST.POLICY_TOP if policy is None else policy
    errors = []

    def parse_functions(f):
        _log.debug(f)
        if not isinstance(f, dict):
            return f
        else:
            (op, value), = f.items()
            if 'equal' == op:
                return value
            elif op in ('greater_or_equal', 'greater_than',
                        'less_than', 'less_or_equal'):
                raise TkStackException(
                    'in parsing {}: function not supported'
                    ''.format(f))
            else:
                raise TkStackException(
                    'in parsing {}: function not recognise'
                    ''.format(f))

    def parse_version(s):
        s = parse_functions(str(s))
        match = re.match('^([0-9]+.)*([0-9]+|x)$', s)
        if match is not None:
            version = re.match('[0-9]+(.[0-9]+)*', s)
            return version.group(0) if version is not None else ''
        else:
            raise TkStackException(
                'in parsing {}: version format is not correct'
                ''.format(s))

    def parse_unit(s):
        # bytes -> bytes
        # kB    -> kilo bytes
        # MB    -> mega bytes
        # GB    -> giga bytes
        s = s.lower()
        if s.endswith('bytes'):
            return int(s[:-5])
        elif s.endswith('kb'):
            return int(s[:-2]) * 1000
        elif s.endswith('mb'):
            return int(s[:-2]) * 1000 * 1000
        elif s.endswith('gb'):
            return int(s[:-2]) * 1000 * 1000 * 1000
        else:
            raise TkStackException(
                'in parsing {}: unit not recognise'.format(s))

    def parse_op(s):
        if s.startswith('>='):
            return 'gte', s[2:]
        elif s.startswith('>'):
            return 'gt', s[1:]
        elif s.startswith('<='):
            return 'lte', s[2:]
        elif s.startswith('<'):
            return 'lt', s[1:]
        else:
            raise TkStackException(
                'in parsing {}: operator not recognise'.format(s))

    def parse_unit_limit(s):
        s = s.strip()
        op, rest = parse_op(s)
        num = parse_unit(rest)
        return op, num

    def parse_limit(s):
        s = s.strip()
        op, rest = parse_op(s)
        return op, int(rest)

    query = {'size_gt': 0}
    if CONST.POLICY_TOP == policy:
        query['sort'] = ('stars', 'pulls', '-size')
    if CONST.POLICY_SIZE == policy:
        query['sort'] = ('-size', 'stars', 'pulls')
    if CONST.POLICY_USED == policy:
        query['sort'] = ('pulls', 'stars', '-size')

    _log.debug('properties {}'.format(properties))
    for p in properties:
        if CONST.PROPERTY_SW in p:
            for s in p[CONST.PROPERTY_SW]:
                (k, v), = s.items()
                try:
                    query[k.lower()] = parse_version(v)
                except TkStackException as e:
                    errors += e.stack
        if CONST.PROPERTY_OS in p:
            try:
                query['distro'] = parse_functions(p[CONST.PROPERTY_OS])
            except TkStackException as e:
                errors += e.stack

    if 'size' in constraints:
        try:
            _log.debug('size {}'.format(constraints['size']))
            op, num = parse_unit_limit(constraints['size'])
            query['size_{}'.format(op)] = num
        except TkStackException as e:
            errors += e.stack
    if 'pulls' in constraints:
        try:
            op, num = parse_limit(constraints['pulls'])
            query['pulls_{}'.format(op)] = num
        except TkStackException as e:
            errors += e.stack
    if 'stars' in constraints:
        try:
            op, num = parse_limit(constraints['stars'])
            query['stars_{}'.format(op)] = num
        except TkStackException as e:
            errors += e.stack

    if len(errors) > 0:
        raise TkStackException(*errors)

    return query


def _choose_image(images, interactive=False):
    def format_size(size):
        return '{}MB'.format(size / 1000 / 1000)

    if interactive:
        # print_('\n')
        print_('{:<3}{:<30}{:>15}{:>15}{:>15}'.format(
               '#', 'NAME', 'SIZE', 'PULLS', 'STARS'))
        for index, image in enumerate(images[:10]):
            print_('{2:<3}{0[name]:<30}{1:>15}'
                   '{0[pulls]:>15}{0[stars]:>15}'.format(
                       image, format_size(image['size']), index + 1))
        # print_('\n')
        i = None
        while not isinstance(i, int):
            try:
                i = int(input('select an image number ')) - 1
                if i < 0 or i > len(images) - 1:
                    _log.debug('not in range')
                    raise Exception()
                return images[i]
            except Exception as e:
                i = None
                print_('must be a number between 1 and {}'.format(len(images)))
    else:
        return images[0] if len(images) > 0 else None


def _update_yaml(node, nodes_yaml, image):
    # update host requirement of all node of the group
    container_name = '{}_container'.format(node.name)
    node_yaml = nodes_yaml[node.name]
    # req_node_yaml = helper.get_host(node_yaml)
    # _log.debug('req_node_yaml {}, type {}'
    #            ''.format(req_node_yaml, type(req_node_yaml)))
    node_host = helper.get_host_node(node_yaml)
    if node_host != container_name:
        _log.debug('there is a different container')
        del nodes_yaml[node_host]
    helper.set_host_node(node_yaml, container_name)

    nodes_yaml[container_name] = _build_container_node(image, node.constraints)


def _update_group_yaml(component, nodes_yaml, image):
    # update host requirement of all node of the group
    container_name = '{}_container'.format(component.name)
    for m in (m for m in component.members if m.to_update):
        node_host = helper.get_host_node(m)
        if node_host != container_name:
            _log.debug('there is a different container')
            del nodes_yaml[node_host]
        helper.set_host_node(nodes_yaml[m.name], container_name)

    # add container node to the template
    node_filter = component.constraints
    nodes_yaml[container_name] = _build_container_node(image, node_filter)


def _build_container_node(image, node_filter):
    def node_filter_to_property(node_filter):
        def list_to_map(l):
            return {k: v for i in l for k, v in i.items()}

        new_prop = {k: copy.deepcopy(list_to_map(v)
                                     if isinstance(v, list) else v)
                    for p in node_filter for k, v in p.items()
                    if CONST.PROPERTY_SW != k and CONST.PROPERTY_OS != k}
        return new_prop

    new_prop = node_filter_to_property(node_filter)\
        if node_filter is not None else {}

    def format_software(softwares):
        return {s['software']: s['ver'] for s in softwares}

    properties = {}
    if 'softwares' in image and len(image['softwares']) > 0:
        properties[CONST.PROPERTY_SW] = format_software(image['softwares'])
    if 'distro' in image and image['distro']:
        properties[CONST.PROPERTY_OS] = image['distro']

    properties.update(new_prop)

    return {
        'type': 'tosker.nodes.Container',
        'properties': properties,
        'artifacts': {
            'my_image': {
                'file': image['name'],
                'type': 'tosker.artifacts.Image',
                'repository': 'docker_hub'
            }
        }
    }



import requests

from .exceptions import TkException
from .helper import CONST, Logger


def search_images(query, df_host):
    try:
        json = _request(df_host, CONST.DF_SEARCH, query)
        return json['count'], json['images']
    except (TkException, KeyError):
        raise TkException('Dockerfinder server error')


def get_software(df_host):
    try:
        json = _request(df_host, CONST.DF_SOFTWARE)
        return [s['name'] for s in json['software']]
    except (TkException, KeyError):
        raise TkException('Dockerfinder software service do not respond')


def _request(df_host, endpoint, params={}):
    _log = Logger.get(__name__)
    url = df_host + endpoint
    try:
        ret = requests.get(url, params=params,
                           headers={'Accept': 'applicaiton/json',
                                    'Content-type': 'application/json'},
                           timeout=5)
        _log.debug('request done on url {}'.format(ret.url))
    except requests.exceptions.RequestException as e:
        _log.debug('Request error: %s', e)
        raise TkException('Request to Dockerfinder fails.')
    return ret.json()



import re

from toscaparser.nodetemplate import NodeTemplate

from . import helper, requester
from .exceptions import TkStackException
from .helper import CONST, Logger


def validate_groups(tosca, groups, force=False):
    # _log = Logger.get(__name__)

    def is_in_members(node, members):
        if isinstance(node, NodeTemplate):
            node = node.name
        return any((node == m.name for m in members))

    errors = []
    all_members = [m.name for g in groups for m in g.members]
    if len(set(all_members)) != len(all_members):
        errors.append('Groups are not disjoint. One or more component apper '
                      'in multiple groups')

    for group in groups:
        # add host to node
        for node in group.members:
            host_node = helper.get_host_node(node)
            node.host = helper.get_node_from_tpl(tosca, host_node)

        # search if same component are hosted on a container
        host_container = next((m.host for m in group.members
                               if m.host is not None and
                               m.host.type == CONST.CONTAINER_TYPE), None)

        for node in group.members:
            # check if all memeber of the group are software
            if node.type != CONST.SOFTWARE_TYPE:
                errors.append(
                    'Group "{}" has a member, "{}", that is not a software '
                    'component'.format(group.name, node.name))
                continue

            # check that if one member of the group has a host container all
            # have to have the same requirement
            if not force and host_container is not None:
                if node.host is None\
                   or (node.host.type == CONST.CONTAINER_TYPE
                       and node.host.name != host_container.name):
                    errors.append(
                        'Group "{0.name}", not all members are hosted on the '
                        '"{1.name}" container (use -f to exclude this check)'
                        ''.format(group, host_container))
                continue

            # check the member of the group do not require host to software
            # nodes outside the group
            if node.host is not None and node.host.type == CONST.SOFTWARE_TYPE:
                if not is_in_members(node.host.name, group.members):
                    errors.append(
                        'Node "{}",  member of the group "{}", requires a '
                        'host to a no member component, "{}"'
                        ''.format(node.name, group.name, node.host.name))

        # check if other software node require host to members of this group
        for node in tosca.nodetemplates:
            # if node is not member of the group
            if all((node.name != m.name for m in group.members)):
                host_node = helper.get_host_node(node)

                # if the node require to be hosted to a member of the group
                if host_node is not None and\
                   is_in_members(host_node, group.members):
                    errors.append(
                        'Node "{}" requires a host to a node member of '
                        'another group "{}"'
                        ''.format(node.name, group.name))
    if len(errors) > 0:
        raise TkStackException(*errors)


def validate_node_filter(tosca, df_host):
    _log = Logger.get(__name__)

    supported_sw = requester.get_software(df_host)
    _log.debug('software on DF {}'.format(supported_sw))

    errors = []
    for node in tosca.nodetemplates:
        node_filter = helper.get_host_nodefilter(node)
        if not isinstance(node_filter, list):
            errors.append('Node "{}": "node_filter" requires a list of '
                          'properties'.format(node.name))
            continue
        for n in node_filter:
            key, value = list(n.items())[0]
            if CONST.PROPERTY_OS == key:
                if not isinstance(value, str):
                    errors.append('Node "{}": property "{}" must be a string'
                                  '.'.format(node.name, CONST.PROPERTY_OS))
            elif CONST.PROPERTY_SW == key:
                if not isinstance(value, list):
                    errors.append('Node "{}": "{}" must be a list of software'
                                  ''.format(node.name, CONST.PROPERTY_SW))
                    continue
                for software in value:
                    s, v = list(software.items())[0]
                    if s not in supported_sw:
                        errors.append(
                            'Node "{}": software "{}" is not supported'
                            ''.format(node.name, s))
                    match = re.match('^([0-9]+.)*([0-9]+|x)?$', str(v))
                    if not isinstance(v, str) or match is None:
                        errors.append(
                            'Node "{}": software version "{}:{}" '
                            'must be a string with this pattern '
                            '([0-9].)*[0-9]+ or ([0-9].)*x '
                            '(i.e. 1.2.2, 1.2.x).'
                            ''.format(node.name, s, v))
            else:
                # TODO: check if other property are container one
                pass

    if len(errors) != 0:
        raise TkStackException(*errors)



class TkException(Exception):

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class TkStackException(Exception):

    def __init__(self, *msg):
        self.stack = list(msg)

    def __str__(self):
        return '- ' + '\n- '.join(self.stack)



import logging
import os
import re
import sys
from copy import copy
from os import path
from sys import argv

from six import StringIO, print_

from . import __version__, toskeriser
from .exceptions import TkException, TkStackException
from .helper import CONST, Logger

_USAGE = '''TosKeriser, a tool to complete TosKer application description with
suitable Docker Images.

toskerise FILE [COMPONENT..] [OPTIONS]
toskerise --supported_sw|-s
toskerise --version|-v
toskerise --help|-h

FILE
  TOSCA YAML file or a CSAR to be completed

COMPONENT
  a list of the components to be completed (by default all component are considered)

OPTIONS
  -f|--force                           force the update of all containers
  --policy=top_rated|size|most_used    ordering of the images
  --constraints=value                  constraint to give to DockerFinder
                                       (e.g. --constraints 'size<=99MB pulls>30
                                                            stars>10')
  -i|--interactive                     active interactive mode
  -q|--quiet                           active quiet mode
  --debug                              active debug mode
'''

_log = None


def run():
    global _log
    if len(argv) > 1:
        try:
            file_path, comps, flags, params = _parse_input(argv[1:])
        except Exception as e:
            print_(''.join(e.args))
            exit(-1)
    else:
        print_('few arguments, --help for usage')
        exit(-1)

    if flags.get('help', False):
        print_(_USAGE)
        exit(0)
    if flags.get('version', False):
        print_('TosKeriser version {}'.format(__version__))
        exit(0)

    df_host = os.environ.get('DOCKERFINDER_HOST', CONST.DF_HOST)
    if flags.get('supported_sw', False):
        software = toskeriser.software_list(df_host)
        print_('Supported software:\n{}'.format('\n'.join(software)))
        exit(0)

    Logger.set_logger(logging.CRITICAL)
    if flags.get('debug', False) and not flags.get('quiet', False):
        Logger.set_logger(logging.DEBUG)
    elif flags.get('quiet', False):
        old_target, sys.stdout = sys.stdout, StringIO()

    _log = Logger.get(__name__)
    _log.debug('DF_HOST: {}'.format(df_host))

    _log.debug('input parameters: {}, {}, {}, {}'.format(
        file_path, comps, flags, params))

    constraint = params.get('constraints', {})
    _log.debug('constraints {}'.format(constraint))

    policy = params.get('policy', None)
    _log.debug('policy {}'.format(policy))

    try:
        toskeriser.toskerise(file_path, components=comps,
                             policy=policy,
                             constraints=constraint,
                             interactive=flags.get('interactive', False),
                             force=flags.get('force', False),
                             df_host=df_host)
    except TkStackException as e:
        print_('ERRORS:\n{}'.format(e))
    except TkException as e:
        print_('Internal error: {}'.format(e))
    finally:
        if flags.get('quiet', False):
            sys.stdout = old_target


def _parse_contraint(con, params=None):
    ret = {}
    size_re = re.compile('(?<=size)\s*(>|<|=|>=|<=)\s*[0-9]+[A-z]+')
    pulls_re = re.compile('(?<=pulls)\s*(>|<|=|>=|<=)\s*[0-9]+')
    stars_re = re.compile('(?<=stars)\s*(>|<|=|>=|<=)\s*[0-9]+')

    size = size_re.search(con)
    if size is not None:
        ret['size'] = size.group(0)

    pulls = pulls_re.search(con)
    if pulls is not None:
        ret['pulls'] = pulls.group(0)

    stars = stars_re.search(con)
    if stars is not None:
        ret['stars'] = stars.group(0)

    return ret


def _parse_policy(policy, params=None):
    if policy is not None and\
       policy not in (CONST.POLICY_TOP, CONST.POLICY_SIZE, CONST.POLICY_USED):
        raise Exception('policy must be "{0.POLICY_TOP}", '
                        '"{0.POLICY_SIZE}" or "{0.POLICY_USED}"'.format(CONST))
    else:
        return policy


_FLAG = {
    '--debug': 'debug',
    '-q': 'quiet',
    '--quiet': 'quiet',
    '--help': 'help',
    '-h': 'help',
    '--supported_sw': 'supported_sw',
    '-s': 'supported_sw',
    '-v': 'version',
    '--version': 'version',
    '-i': 'interactive',
    '--interactive': 'interactive',
    '-f': 'force',
    '--force': 'force'
}

_PARAMS = {
    '--policy': _parse_policy,
    '--constraints': _parse_contraint,
}


def _parse_input(args):
    params = {}
    comps = []
    flags = {}
    file = ''
    p1 = re.compile('--.*')
    p2 = re.compile('-.??$')
    p3 = re.compile('-.*')

    def check_file(file):
        file_name = None
        if path.isfile(file):
            if file.endswith(('.yaml', '.csar', '.CSAR', '.YAML')):
                file_name = file
        return file_name

    def get_value(i):
        old_i = i
        while i < len(args) and (not p1.match(args[i]) and
                                 not p2.match(args[i])):
            i += 1
        return ' '.join(args[old_i:i]), i

    i = 0
    while i < len(args):
        if p1.match(args[i]):
            if _FLAG.get(args[i], False):
                flags[_FLAG[args[i]]] = True
            elif args[i] in _PARAMS:
                if (i + 1 < len(args) and
                    not p1.match(args[i + 1]) and
                        not p2.match(args[i + 1])):
                    value, new_i = get_value(i + 1)
                    params[args[i][2:]] = _PARAMS[args[i]](
                        value,
                        copy(params.get(args[i][2:], None)))
                    i = new_i
                    continue
                else:
                    raise Exception('missing input value for {}'.format(
                        args[i]))
            else:
                raise Exception('parameter {} not recognise'.format(args[i]))
        elif p2.match(args[i]):
            if _FLAG.get(args[i], False):
                flags[_FLAG[args[i]]] = True
            else:
                raise Exception('known parameter.')
        elif p3.match(args[i]):
            for c in args[i][1:]:
                c = '-{}'.format(c)
                if _FLAG.get(c, False):
                    flags[_FLAG[c]] = True
                else:
                    raise Exception('known parameter.')
        elif args[i] and file:
            comps.append(args[i])
        elif i == 0:
            file = check_file(args[i])
            if file is None:
                raise Exception('first argument must be a TOSCA yaml file, '
                                'a CSAR or a ZIP archive.')
        i += 1
    return file, comps, flags, params


if __name__ == '__main__':
    run()



from .exceptions import TkStackException
from .helper import CONST, Logger

_log = None


def merge_constraint(nodes_property):
    global _log
    _log = Logger.get(__name__)

    merged_properties = {}
    errors = []
    for n in nodes_property:
        for p in n:
            key, value = list(p.items())[0]
            _log.debug('k:{} v:{}'.format(key, value))

            try:
                if CONST.PROPERTY_SW == key:
                    _add_property(merged_properties, p, f_merge=_merge_version)
                else:
                    # if CONST.PROPERTY_SW of other properties
                    _add_property(merged_properties, p)
            except TkStackException as e:
                errors += e.stack

    if len(errors) != 0:
        raise TkStackException(*errors)
    _log.debug(merged_properties)
    p_list = _convert_to_list(merged_properties)
    _log.debug(p_list)
    return p_list


def _add_property(merged_p, new_p, f_merge=lambda x, y: x if x == y else None):
    '''
    Add a property to a dictionary of properties executing the f_merge to
    resolve conflicts.
    '''
    new_p_key, new_p_value = list(new_p.items())[0]

    errors = []
    # if the property to merge is a list
    if isinstance(new_p_value, list):
        if new_p_key not in merged_p:
            merged_p[new_p_key] = {}

        for p in new_p_value:
            p_name, p_value = list(p.items())[0]

            if p_name not in merged_p[new_p_key]:
                merged_p[new_p_key][p_name] = p_value
            else:
                res = f_merge(merged_p[new_p_key][p_name], p_value)
                if res is not None:
                    merged_p[new_p_key][p_name] = res
                else:
                    errors.append('Cannot merge value "{}" with "{}" of '
                                  'property "{}"'
                                  ''.format(
                                      merged_p[new_p_key][p_name],
                                      p_value, new_p_key + '.' + str(p_name)))

    # if the property to merge is a string
    else:
        if new_p_key not in merged_p:
            merged_p[new_p_key] = new_p_value
        else:
            res = f_merge(merged_p[new_p_key], new_p_value)
            if res is not None:
                merged_p[new_p_key] = res
            else:
                errors.append('Cannot merge value "{}" with "{}" of '
                              'property "{}"'
                              ''.format(merged_p[new_p_key], new_p_value,
                                        new_p_key))

    if len(errors) != 0:
        raise TkStackException(*errors)


def _merge_version(v1, v2):
    '''
    Merge two software version. It return the merged one or None
    if the merge is not possible
    '''
    v1 = v1.split('.')
    v2 = v2.split('.')
    min_len = len(v1) if len(v1) < len(v2) else len(v2)

    merged_version = []
    i = 0
    while i < min_len and v1[i] != 'x' and v2[i] != 'x':
        if v1[i] == v2[i]:
            merged_version.append(v1[i])
        else:
            return None
        i += 1

    if i < len(v1) and v1[i] == 'x':
        return '.'.join(merged_version + v2[i:])
    elif i < len(v2) and v2[i] == 'x':
        return '.'.join(merged_version + v1[i:])
    elif len(v1) != len(v2):
        return None
    else:
        return '.'.join(merged_version)


def _convert_to_list(properties):
    if isinstance(properties, dict) and 'get_input' not in properties.keys():
        return [{k: _convert_to_list(v)} for k, v in properties.items()]
    else:
        return properties



import logging
import os
import zipfile
from os import path

from six import string_types
from toscaparser.prereq.csar import CSAR


class CONST:
    # DF_HOST = 'http://131.114.2.77/search'
    # DF_HOST = 'http://127.0.0.1:3000/search'
    DF_HOST = 'http://black.di.unipi.it'
    DF_SEARCH = ':3000/search'
    DF_SOFTWARE = ':3001/api/software'

    CONTAINER_TYPE = 'tosker.nodes.Container'
    SOFTWARE_TYPE = 'tosker.nodes.Software'
    # IMAGE_TYPE = 'tosker.artifacts.Image'
    # DOCKERFILE_TYPE = 'tosker.artifacts.Dockerfile'

    PROPERTY_SW = 'supported_sw'
    PROPERTY_OS = 'os_distribution'
    REQUIREMENT_HOST = 'host'

    POLICY_TOP = 'top_rated'
    POLICY_SIZE = 'size'
    POLICY_USED = 'most_used'


class Logger:
    _ch = None

    @staticmethod
    def set_logger(level=logging.DEBUG):
        # global Logger._ch
        Logger._ch = logging.StreamHandler()
        Logger._ch.setLevel(level)
        formatter = logging.Formatter((
            '%(levelname) -3s %(asctime)s %(name)'
            '-3s %(funcName)'
            '-1s %(lineno) -0s: %(message)s'
        ))
        Logger._ch.setFormatter(formatter)

    @staticmethod
    def get(name, level=logging.DEBUG):
        log = logging.getLogger(name)
        log.setLevel(level)
        if Logger._ch is not None:
            log.addHandler(Logger._ch)
        return log


class Group:

    def __init__(self, name, members):
        self.name = name
        self.members = members


def unpack_csar(file_path):
    # Work around bug validation csar of toscaparser
    csar = CSAR(file_path)
    try:
        csar.validate()
    except ValueError as e:
        # _log.debug(e)
        if not str(e).startswith("The resource") or \
           not str(e).endswith("does not exist."):
            raise e

    csar.decompress()

    yaml_path = path.join(csar.temp_dir, csar.get_main_template())
    return (csar.temp_dir, yaml_path)


def pack_csar(csar_tmp_path, new_path):
    with zipfile.ZipFile(new_path, "w", zipfile.ZIP_DEFLATED) as zf:
        for dirname, subdirs, files in os.walk(csar_tmp_path):
            # zf.write(dirname, root)
            for filename in files:
                zf.write(path.join(dirname, filename),
                         path.relpath(path.join(dirname, filename),
                                      path.join(csar_tmp_path, '.')))


def convert_tosca_group(tosca):
    return [Group(g.name,
                  [get_node_from_tpl(tosca, m) for m in g.members])
            for g in tosca.topology_template.groups]


def get_host(node):
    if hasattr(node, 'entity_tpl'):
        node = node.entity_tpl
    if 'requirements' in node:
        for r in node['requirements']:
            (k, v), = r.items()
            if CONST.REQUIREMENT_HOST == k:
                if v is None:
                    r[k] = {}
                return r[k]
    return None


def get_host_key(node, key):
    requirement = get_host(node)
    if requirement is not None and isinstance(requirement, dict):
        if key in requirement:
            return requirement[key]
    return None


def set_host_node(node, host_node):
    if 'requirements' in node:
        for r in node['requirements']:
            (k, v), = r.items()
            if CONST.REQUIREMENT_HOST == k:
                if not isinstance(v, dict):
                    r[k] = host_node
                else:
                    r[k]['node'] = host_node
                return True
    return False


def get_host_node(node):
    requirement = get_host(node)
    if isinstance(requirement, string_types):
        return requirement
    return get_host_key(node, 'node')


def get_host_nodefilter(node):
    requirement = get_host(node)
    try:
        properties = requirement['node_filter']['properties'] or []
    except (TypeError,  KeyError):
        properties = []
    return properties


def get_node_from_tpl(tosca, str_node):
    for node in tosca.nodetemplates:
        if node.name == str_node:
            return node
    return None



# Always prefer setuptools over distutils
# To use a consistent encoding
from codecs import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

exec(open('toskeriser/__init__.py').read())

setup(
    name='TosKeriser',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=__version__,

    description='A tool to complete TosKer application description with'
                'suitable Docker Images',

    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/di-unipi-socc/TosKeriser',

    # Author details
    author='lucarin91',
    author_email='to@lucar.in',

    # Choose your license
    license='MIT',

    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: System :: Installation/Setup',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='Docker match matcher TOSCA deployment complete development',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    # packages=find_packages(exclude=['test']),
    packages=find_packages(exclude=['tests']),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['cmd2<=0.8.6',  # BUGFIX: Added to avoide conflict in the
                                      # version of cmd2 inside tosca_parser.
                      'tosca-parser', 'six', 'requests', 'ruamel.yaml'],


    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest', 'isort', 'flake8', 'tox', 'coverage'],
        'test': ['requests_mock'],
    },

    test_suite="tests",

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    include_package_data=True,
    # package_data={
    #     'tosker': ['*.yaml'],
    # },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[
        ('/usr/share/toskeriser', ['data/tosker-types.yaml']),
        ('/usr/share/toskeriser/examples',
            ['data/examples/thinking-app/thinking.csar'])

    ],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'toskerise=toskeriser.ui:run',
        ],
    },
)



import traceback

import ruamel.yaml
# from six import print_
from toscaparser.common.exception import ValidationError
from toscaparser.tosca_template import ToscaTemplate

from . import completer, helper, merger, requester, validator
from .exceptions import TkException, TkStackException
from .helper import CONST, Logger

_log = None


def toskerise(file_path, components=[], policy=None, constraints={},
              interactive=False, force=False, df_host=CONST.DF_HOST):
    global _log
    _log = Logger.get(__name__)

    is_csar = file_path.endswith(('.zip', '.csar', '.CSAR'))
    _log.debug('CSAR founded: {}'.format(is_csar))

    def get_new_path(partial):
        def gen_new_path(mod):
            points = file_path.split('.')
            return '{}.{}.{}'.format('.'.join(points[:-1]), mod, points[-1])
        return gen_new_path('partial') \
            if partial else\
            gen_new_path('completed')

    try:
        if is_csar:
            # extract CSAR
            csar_tmp_path, yaml_path = helper.unpack_csar(file_path)
            # process TOSCA
            is_partial, tosca_yaml = _process_tosca(yaml_path, components,
                                                    policy, constraints,
                                                    interactive, force,
                                                    df_host)
            # write changes
            _write_updates(tosca_yaml, yaml_path)
            # pack CSAR
            helper.pack_csar(csar_tmp_path, get_new_path(is_partial))
        else:
            # process TOSCA
            is_partial, tosca_yaml = _process_tosca(file_path, components,
                                                    policy, constraints,
                                                    interactive, force,
                                                    df_host)
            # write changes
            _write_updates(tosca_yaml, get_new_path(is_partial))
    except ValidationError as e:
        raise TkStackException('Validation error:{}'.format(e))
    except TkStackException as e:
        raise e
    except Exception as e:
        _log.debug(traceback.format_exc())
        raise TkException('Internal error: {}'.format(e))


def software_list(df_host):
    return sorted(requester.get_software(df_host))


def _process_tosca(file_path, components=[], policy=None, constraints={},
                   interactive=False, force=False, df_host=CONST.DF_HOST):
    # parse TOSCA
    tosca = ToscaTemplate(file_path)

    # Check TOSCA dependent input
    _check_components(tosca, components)

    # validation
    validator.validate_node_filter(tosca, df_host)
    groups = helper.convert_tosca_group(tosca)
    validator.validate_groups(tosca, groups, force)

    # filter and merge
    to_complete = _filter_and_merge(tosca, groups, force, components)
    if len(to_complete) == 0:
        raise TkStackException('no abstract node founded')

    # get yaml to be updated
    tosca_yaml = _get_roundtrip_node(file_path)
    node_yaml = tosca_yaml['topology_template']['node_templates']

    errors = []
    all_completed = True
    for component in to_complete:
        try:
            _log.debug('start completation of {}'.format(component.name))
            all_completed &= completer.complete(component, node_yaml, tosca,
                                                policy, constraints,
                                                interactive, df_host)
        except TkStackException as e:
            errors += e.stack

    if len(errors) > 0:
        raise TkStackException(*errors)

    return not all_completed, tosca_yaml


def _filter_and_merge(tosca, groups, force, components):
    '''
    Filter the groups and the nodes and returning a list of components
    to be update.
    '''
    already_analysed = []
    errors = []
    to_update = []

    for group in groups:
        constraints = []

        group_to_complete = False
        for m in group.members:
            m.to_update = _must_update(tosca, m, force, components)
            group_to_complete |= m.to_update
            already_analysed.append(m.name)
            constraints.append(helper.get_host_nodefilter(m))

        if group_to_complete:
            try:
                group.constraints = merger.merge_constraint(constraints)
                _log.debug('merged properties {}'.format(group.constraints))
                to_update.append(group)
            except TkStackException as e:
                errors += e.stack

    # remove the node that are already processed as part of a group
    for node in (n for n in tosca.nodetemplates
                 if n.name not in already_analysed):
        if _must_update(tosca, node, force, components):
            _log.debug('node {.name} is abstract'.format(node))
            node.constraints = helper.get_host_nodefilter(node)
            to_update.append(node)

    if len(errors) != 0:
        raise TkStackException(*errors)
    return to_update


def _must_update(tosca, node, force, components):
    '''
    Check if a node has to be update.
    '''
    def is_software(node):
        return True if node.type == CONST.SOFTWARE_TYPE else False

    def has_node(node):
        return helper.get_host_node(node) is not None

    def is_bottom(node):
        host_node = helper.get_node_from_tpl(
                        tosca,
                        helper.get_host_node(node)
                    )
        return host_node is None or not is_software(host_node)

    if is_software(node)\
       and is_bottom(node)\
       and (len(components) == 0 or node.name in components)\
       and (not has_node(node) or force):
        return True
    return False


def _get_roundtrip_node(file_path):
    with open(file_path, 'r') as f:
        tosca_yaml = ruamel.yaml.round_trip_load(f, preserve_quotes=True)
    return tosca_yaml


def _write_updates(tosca, new_path):
    with open(new_path, 'w') as f:
        ruamel.yaml.round_trip_dump(tosca, f,
                                    # width=80,
                                    # indent=2,
                                    # block_seq_indent=2,
                                    # top_level_colon_align=True
                                    )


def _check_components(tosca, components):
    if hasattr(tosca, 'nodetemplates'):
        if tosca.nodetemplates:
            for c in components:
                correct = False
                for n in tosca.nodetemplates:
                    if c == n.name:
                        correct = True
                if not correct:
                    raise TkStackException(
                        'component "{}" not founded'.format(c))
